from pathlib import Path
from typing import Any, Dict, List, Union, Optional, Callable

from pyspark.ml.param import Param
from pyspark.sql.types import StructType

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_local_file_loader.v1.framework_local_file_loader import (
    FrameworkLocalFileLoader,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from spark_pipeline_framework.utilities.file_modes import FileReadModes


class FrameworkCsvLoader(FrameworkLocalFileLoader):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        file_path: Union[
            str, List[str], Path, Callable[[Optional[str]], Union[Path, str]]
        ],
        delimiter: str = ",",
        has_header: bool = True,
        clean_column_names: bool = False,
        multiline: bool = False,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        limit: Optional[int] = None,
        infer_schema: bool = False,
        cache_table: bool = True,
        schema: Optional[StructType] = None,
        create_file_path: bool = False,
        mapping_file_name: Optional[str] = None,
        mode: str = FileReadModes.MODE_PERMISSIVE,
        stream: bool = False,
        delta_lake_table: Optional[str] = None,
    ) -> None:
        """
        class that handles loading of csv files


        :param view: view to populate with the data in the file
        :param file_path: where to store the FHIR resources retrieved.  Can be either a path or a function
                    that returns path. Function is passed a loop number if inside a loop or else None
        :param delimiter:
        :param limit:
        :param has_header:
        :param infer_schema:
        :param cache_table: whether to cache the table in memory after loading it
        :param schema: schema to use when loading the data
        :param clean_column_names: whether to remove invalid characters from column names
                                    (i.e., characters that Spark does not like)
        :param create_file_path: whether to create the file path if it does not exist
        :param mode: file loading mode
        :param stream: whether to stream the data
        :param delta_lake_table: store data in this delta lake
        """

        super().__init__(
            view=view,
            file_path=file_path,
            clean_column_names=clean_column_names,
            name=name,
            parameters=parameters,
            progress_logger=progress_logger,
            delimiter=delimiter,
            limit=limit,
            has_header=has_header,
            infer_schema=infer_schema,
            cache_table=cache_table,
            schema=schema,
            create_file_path=create_file_path,
            mode=mode,
            stream=stream,
            delta_lake_table=delta_lake_table,
        )

        self.multiline: Param[bool] = Param(self, "multiline", "")
        self._setDefault(multiline=multiline)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDelimiter(self) -> str:
        return self.getOrDefault(self.delimiter)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getHasHeader(self) -> bool:
        return self.getOrDefault(self.has_header)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMultiline(self) -> bool:
        return self.getOrDefault(self.multiline)

    def getReaderFormat(self) -> str:
        return "csv"

    def getReaderOptions(self) -> Dict[str, Any]:
        options: Dict[str, Any] = {
            "header": "true" if self.getHasHeader() else "false",
            "delimiter": self.getDelimiter(),
            "multiline": self.getMultiline(),
        }
        return options



import json
from typing import Any, Dict, List, Optional, TYPE_CHECKING, TypeVar

# noinspection PyPackageRequirements
from pyspark.ml.base import Transformer

# noinspection PyPackageRequirements
from pyspark.ml.param import Param

# noinspection PyPackageRequirements
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable

# noinspection PyPackageRequirements
from pyspark.sql.dataframe import DataFrame
from typing_extensions import final

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.utilities.class_helpers import ClassHelpers

T = TypeVar("T")


class FrameworkTransformer(
    Transformer,
    DefaultParamsReadable,  # type: ignore
    DefaultParamsWritable,
):
    # noinspection PyUnusedLocal
    def __init__(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        Base class for our transformers

        """
        super().__init__()

        if TYPE_CHECKING:
            self._input_kwargs: Dict[str, Any] = {}
        self.logger = get_logger(__name__)

        self.name: Param[str] = Param(self, "name", "")
        self._setDefault(name=name)

        self.progress_logger: Param[Optional[ProgressLogger]] = Param(
            self, "progress_logger", ""
        )
        self._setDefault(progress_logger=progress_logger)

        self.loop_id: Optional[str] = None

        self.parameters: Optional[Dict[str, Any]] = parameters

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal
    def setStandardParams(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> Any:
        kwargs: Dict[str, Any] = self._input_kwargs
        return self._set(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal,Mypy
    @final
    def setParams(self, **kwargs: Any) -> Any:
        # ignore any parameters
        kwargs = {key: value for key, value in kwargs.items() if self.hasParam(key)}
        return self._set(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        return df

    @property
    def transformers(self) -> List[Transformer]:
        return [self]

    # # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    # def setName(self, value: str) -> "FrameworkTransformer":
    #     self._paramMap[self.name] = value
    #     return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return self.getOrDefault(self.name)

    # # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    # def setProgressLogger(
    #     self, value: Optional[ProgressLogger]
    # ) -> "FrameworkTransformer":
    #     self._paramMap[self.progress_logger] = value
    #     return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getProgressLogger(self) -> Optional[ProgressLogger]:
        return self.getOrDefault(self.progress_logger)

    # # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    # def setParameters(self, value: Optional[Dict[str, Any]]) -> "FrameworkTransformer":
    #     self._paramMap[self.parameters] = value
    #     return self
    #
    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getParameters(self) -> Optional[Dict[str, Any]]:
        return self.parameters

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSql(self) -> Optional[str]:
        return None

    # # This is here to avoid mypy from complaining since this is a protected member
    # noinspection PyPep8Naming
    def _setDefault(self, **kwargs: Any) -> None:
        # noinspection PyUnresolvedReferences,PyProtectedMember
        super()._setDefault(**kwargs)

    def _set(self, **kwargs: Any) -> None:
        # filter out any args that don't have parameters
        kwargs = {key: value for key, value in kwargs.items() if self.hasParam(key)}
        # noinspection PyUnresolvedReferences,PyProtectedMember
        super()._set(**kwargs)

    def __str__(self) -> str:
        return json.dumps(self.as_dict(), default=str)

    def as_dict(self) -> Dict[str, Any]:
        return {
            "name": self.getName(),
            "short_type": self.__class__.__name__,
            "type": ClassHelpers.get_full_name_of_instance(self),
            "params": {
                k.name: (
                    self.getOrDefault(k)
                    if not hasattr(self.getOrDefault(k), "as_dict")
                    else self.getOrDefault(k).as_dict()
                )
                if not callable(self.getOrDefault(k))
                else ClassHelpers.get_function_as_text(
                    fn=self.getOrDefault(k), strip=f"{k.name}="
                )
                for k, v in self._paramMap.items()
                if k.name not in ["progress_logger"]
            },
        }

    def update_from_dict(self, dictionary: Dict[str, Any]) -> "FrameworkTransformer":
        if dictionary:
            for key, value in dictionary.items():
                setattr(self, key, value)
        return self

    # Have to re-declare here because MyPy does not seem to pick up the overload from base class (Params)
    def getOrDefault(self, param: Param[T]) -> T:
        # noinspection PyUnresolvedReferences
        # return super(Params, self).getOrDefault(param)  # type: ignore
        return super().getOrDefault(param)  # type: ignore

    def set_loop_id(self, loop_id: str) -> None:
        """
        Set when running inside a FrameworkLoopTransformer

        :param loop_id: loop id
        """
        self.loop_id = loop_id



import pkgutil
import re
import sys
import types
from importlib import import_module
from inspect import signature
from typing import Dict, Any, Optional, Callable, Union, List, cast

from pyspark.ml import Transformer
from spark_auto_mapper.automappers.automapper_base import AutoMapperBase

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger


def get_python_transformer_from_location(
    location: str,
    import_module_name: str,
    parameters: Dict[str, Any],
    progress_logger: Optional[ProgressLogger],
    mapping_file_name: Optional[str] = None,
) -> Transformer:
    assert location
    # assert isinstance(parameters, dict)
    assert progress_logger
    search = re.search(r"/library/", location)
    assert search
    start_ = search.start() + 1
    lib_path = location[start_:].replace("/", ".").replace("", "")
    # load_all_modules_from_dir(location)
    module = import_module(import_module_name, lib_path)
    md = module.__dict__
    my_class = [
        md[c]
        for c in md
        if (isinstance(md[c], type) and md[c].__module__ == module.__name__)
    ][0]
    my_class_signature = signature(my_class.__init__)
    my_class_args = [
        param.name
        for param in my_class_signature.parameters.values()
        if param.name != "self"
    ]
    # now figure out the class_parameters to use when instantiating the class
    class_parameters = parameters.copy()
    class_parameters["parameters"] = parameters
    class_parameters["progress_logger"] = progress_logger
    class_parameters["mapping_file_name"] = mapping_file_name
    if len(my_class_args) > 0 and len(class_parameters) > 0:
        return cast(
            Transformer,
            my_class(
                **{k: v for k, v in class_parameters.items() if k in my_class_args}
            ),
        )
    else:
        return cast(Transformer, my_class())


def get_python_function_from_location(
    location: str, import_module_name: str, function_name: Optional[str] = None
) -> Callable[[Dict[str, Any]], Union[AutoMapperBase, List[AutoMapperBase]]]:
    assert location
    search = re.search(r"/library/", location)
    assert search
    start_ = search.start() + 1
    lib_path = location[start_:].replace("/", ".").replace("", "")
    # load_all_modules_from_dir(location)
    module = import_module(import_module_name, lib_path)
    md = module.__dict__
    # noinspection PyTypeChecker
    my_functions: List[
        Callable[[Dict[str, Any]], Union[AutoMapperBase, List[AutoMapperBase]]]
    ] = (
        [md[c] for c in md if isinstance(md[c], types.FunctionType)]
        if not function_name
        else [
            md[c]
            for c in md
            if isinstance(md[c], types.FunctionType) and c == function_name
        ]
    )
    my_function: Callable[
        [Dict[str, Any]], Union[AutoMapperBase, List[AutoMapperBase]]
    ] = my_functions[0]
    my_function_signature = signature(my_function)
    my_function_args = [
        param
        for param in my_function_signature.parameters.values()
        if param.name != "self"
    ]
    # now figure out the function_parameters has one and only one parameter: parameters
    assert len(my_function_args) == 1
    assert my_function_args[0].name == "parameters"
    return my_function


def load_all_modules_from_dir(dirname: str) -> None:
    for importer, package_name, _ in pkgutil.iter_modules([dirname]):
        full_package_name = "%s.%s" % (dirname, package_name)
        if full_package_name not in sys.modules:
            module = importer.find_module(package_name).load_module(full_package_name)  # type: ignore
            print(module)



from pathlib import Path
from typing import Dict, Any, Optional, Union, Callable

from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_parquet_exporter.v1.framework_parquet_exporter import (
    FrameworkParquetExporter,
)
from spark_pipeline_framework.transformers.framework_parquet_loader.v1.framework_parquet_loader import (
    FrameworkParquetLoader,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters


class FrameworkCheckpoint(FrameworkTransformer):
    """
    Saves given view to parquet and reloads it
    """

    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]],
        view: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        stream: bool = False,
        delta_lake_table: Optional[str] = None,
    ):
        """
        Saves given view to parquet and reload it.
        This is to cut off the Spark lineage by saving to temporary location thus making Spark faster in some cases
        https://jaceklaskowski.gitbooks.io/mastering-spark-sql/content/spark-sql-checkpointing.html
        Normally Spark will run from the beginning of the pipeline to the end but in complex pipelines this can cause
        slowness and "out of memory" errors.
        By doing a checkpoint, we can improve this.  However, the downside of checkpointing is that you can slow the
        pipeline since you're writing out a temporary result to the disk.  So this should be used ONLY when the
        pipelines are slow or running out of memory.


        :param view: view to save to parquet
        :param file_path: where to save
        :param name: a name for the transformer step
        :param parameters: parameters
        :param progress_logger: the logger to use for logging
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        assert (
            isinstance(file_path, Path)
            or isinstance(file_path, str)
            or callable(file_path)
        ), type(file_path)
        assert file_path

        assert view

        # add a param
        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.file_path: Param[Union[str, Path]] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.stream: Param[bool] = Param(self, "stream", "")
        self._setDefault(stream=stream)

        self.delta_lake_table: Param[Optional[str]] = Param(
            self, "delta_lake_table", ""
        )
        self._setDefault(delta_lake_table=delta_lake_table)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        file_path: Union[
            Path, str, Callable[[Optional[str]], Union[Path, str]]
        ] = self.getFilePath()
        if callable(file_path):
            file_path = file_path(self.loop_id)
        stream: bool = self.getStream()

        save_transformer = FrameworkParquetExporter(
            view=view,
            name=f"{self.getName()}-save",
            file_path=file_path,
            parameters=self.getParameters(),
            progress_logger=self.getProgressLogger(),
            stream=stream,
        )
        df = save_transformer.transform(df)

        load_transformer = FrameworkParquetLoader(
            view=view,
            file_path=file_path,
            name=f"{self.getName()}-load",
            parameters=self.getParameters(),
            progress_logger=self.getProgressLogger(),
            stream=stream,
        )
        df = load_transformer.transform(df)
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getStream(self) -> bool:
        return self.getOrDefault(self.stream)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDeltaLakeTable(self) -> Optional[str]:
        return self.getOrDefault(self.delta_lake_table)



from typing import Optional

from spark_pipeline_framework.event_loggers.event_logger import EventLogger
from spark_pipeline_framework.utilities.slack.base_slack_client import BaseSlackClient

from spark_pipeline_framework.utilities.slack.slack_client import SlackClient
from spark_pipeline_framework.utilities.slack.slack_client_native import (
    SlackClientNative,
)


class SlackEventLogger(EventLogger):
    def __init__(
        self,
        id_: str,
        slack_token: str,
        slack_channel: str,
        bot_user_name: str,
        slack_error_channel: Optional[str] = None,
        log_placeholder_url: Optional[str] = None,
        flow_run_name: Optional[str] = None,
        backoff: bool = True,
        use_native_slack_client: bool = False,
    ) -> None:
        """
        Logs events to Slack

        :param id_: unique id for this logger
        :param slack_token: token to auth with Slack
        :param slack_channel: channel to post message into
        :param slack_error_channel: channel to post error message into
        :param bot_user_name: user name to use when posting messages
        :param log_placeholder_url: url to logs with a placeholder for flow_run_name to be replaced by id
                e.g., https://grafana.services.icanbwell.com/explore?orgId=1&left=%5B%22now-1h%22,%22
                        now%22,%22loki.staging.icanbwell.com%22,%7B%22expr%22:%22%7Bflow_run_name%3D%5C%22
                        {flow_run_name}%5C%22%7D%22%7D%5D
        :param backoff: whether to backoff sending too many messages to slack
        :param flow_run_name: flow run name to set in log_placeholder_url
        :param use_native_slack_client: whether to use the native slack client or use plain HTTP post
        """
        super().__init__()
        assert slack_token
        assert slack_channel
        assert bot_user_name
        self.id_: str = id_
        self.slack_token: str = slack_token
        self.slack_channel: str = slack_channel
        self.slack_error_channel: Optional[str] = slack_error_channel
        self.slack_user_name: str = bot_user_name
        self.backoff: bool = backoff
        self.slack_client: BaseSlackClient = (
            SlackClientNative(
                slack_token=slack_token,
                channel=slack_channel,
                bot_user_name=bot_user_name,
            )
            if use_native_slack_client
            else SlackClient(
                slack_token=slack_token,
                channel=slack_channel,
                bot_user_name=bot_user_name,
            )
        )
        self.slack_error_client: Optional[BaseSlackClient] = (
            (
                SlackClientNative(
                    slack_token=slack_token,
                    channel=slack_error_channel,
                    bot_user_name=bot_user_name,
                )
                if use_native_slack_client
                else SlackClient(
                    slack_token=slack_token,
                    channel=slack_error_channel,
                    bot_user_name=bot_user_name,
                )
            )
            if slack_error_channel
            else None
        )
        self.flow_run_name: Optional[str] = flow_run_name
        self.log_placeholder_url: Optional[str] = log_placeholder_url
        self.has_sent_log_url: bool = False

    def log_progress_event(
        self,
        event_name: str,
        current: int,
        total: int,
        event_format_string: str,
        backoff: bool = True,
    ) -> None:
        # for first 10 batches send every time
        # for 10 to 100 send every 10 batches
        # after 100 send every 100 batches
        if (
            not self.backoff
            or not backoff
            or (current < 10)
            or (current < 100 and current % 10 == 0)
            or (current < 1000 and current % 100 == 0)
            or (current < 10000 and current % 1000 == 0)
            or (current < 100000 and current % 10000 == 0)
            or (current < 1000000 and current % 100000 == 0)
        ):
            self.slack_client.post_message_to_slack(
                event_format_string.format(event_name, current, total)
            )

    def log_event(self, event_name: str, event_text: str) -> None:
        self.slack_client.post_message_to_slack(event_text)
        if not self.has_sent_log_url:
            if self.log_placeholder_url:
                log_url = self.get_grafana_url()
                if log_url:
                    self.slack_client.post_message_to_slack(log_url)
            self.has_sent_log_url = True

    def log_exception(self, event_name: str, event_text: str, ex: Exception) -> None:
        # don't send full exception to slack since it can have PHI
        self.slack_client.post_message_to_slack(
            f"Helix Pipeline Failure: {self.id_} {event_name}: {event_text} {type(ex)}. <{self.get_grafana_url()}>"
        )

        if self.slack_error_client:
            self.slack_error_client.post_message_to_slack(
                f"Helix Pipeline Failure: {self.id_} {event_name}: {event_text} {type(ex)}. <{self.get_grafana_url()}>"
            )

    def get_grafana_url(self) -> Optional[str]:
        return (
            self.log_placeholder_url.format(flow_run_name=f"{self.flow_run_name}")
            if self.log_placeholder_url
            else None
        )



from pathlib import Path

import pytest
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import (
    FrameworkCsvLoader,
)
from spark_pipeline_framework.transformers.framework_validation_transformer.v1.framework_validation_transformer import (
    FrameworkValidationTransformer,
)

from tests.conftest import clean_spark_session


def test_validation_throws_error(spark_session: SparkSession) -> None:
    with pytest.raises(AssertionError):
        clean_spark_session(spark_session)
        query_dir: Path = Path(__file__).parent.joinpath("./queries")
        data_dir: Path = Path(__file__).parent.joinpath("./data")
        test_data_file: str = f"{data_dir.joinpath('test.csv')}"
        validation_query_file: str = "validate.sql"

        schema = StructType([])

        df: DataFrame = spark_session.createDataFrame(
            spark_session.sparkContext.emptyRDD(), schema
        )

        FrameworkCsvLoader(view="my_view", file_path=test_data_file).transform(df)

        FrameworkValidationTransformer(
            validation_source_path=str(query_dir),
            validation_queries=[validation_query_file],
            fail_on_validation=True,
        ).transform(df)

        df.sql_ctx.table("pipeline_validation").show(truncate=False)


def test_validation_records_error(spark_session: SparkSession) -> None:
    clean_spark_session(spark_session)
    query_dir: Path = Path(__file__).parent.joinpath("./queries")
    data_dir: Path = Path(__file__).parent.joinpath("./data")
    test_data_file: str = f"{data_dir.joinpath('test.csv')}"
    validation_query_file: str = "validate.sql"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    FrameworkCsvLoader(view="my_view", file_path=test_data_file).transform(df)

    FrameworkValidationTransformer(
        validation_source_path=str(query_dir),
        validation_queries=[validation_query_file],
    ).transform(df)

    df_validation = df.sql_ctx.table("pipeline_validation")
    assert (
        1 == df_validation.filter("is_failed == 1").count()
    ), "Expected one failing row in the validation table"


def test_validation_recurses_query_dir(spark_session: SparkSession) -> None:
    clean_spark_session(spark_session)
    query_dir: Path = Path(__file__).parent.joinpath("./queries")
    more_queries_dir: str = "more_queries"
    data_dir: Path = Path(__file__).parent.joinpath("./data")
    test_data_file: str = f"{data_dir.joinpath('test.csv')}"
    validation_query_file: str = "validate.sql"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    FrameworkCsvLoader(view="my_view", file_path=test_data_file).transform(df)

    FrameworkValidationTransformer(
        validation_source_path=str(query_dir),
        validation_queries=[validation_query_file, more_queries_dir],
    ).transform(df)

    df_validation = df.sql_ctx.table("pipeline_validation")
    df_validation.show(truncate=False)
    assert 3 == df_validation.count(), "Expected 3 total rows in pipeline_validation"
    assert (
        1 == df_validation.filter("is_failed == 1").count()
    ), "Expected one failing row in the validation table"



from typing import Any, Callable, Dict, List, Optional, Union

# noinspection PyProtectedMember
from mlflow.entities import RunStatus  # type: ignore
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql import DataFrame
from spark_auto_mapper.automappers.automapper import AutoMapper
from spark_auto_mapper.automappers.automapper_base import AutoMapperBase

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_mapping_runner.v1.framework_mapping_runner_exception import (
    FrameworkMappingRunnerException,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)

# define type for AutoMapperFunction
AutoMapperTypeOrList = Union[AutoMapperBase, List[AutoMapperBase]]
AutoMapperFunction = Callable[[Dict[str, Any]], AutoMapperTypeOrList]


class FrameworkMappingLoader(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        mapping_function: AutoMapperFunction,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        mapping_file_name: Optional[str] = None,
    ) -> None:
        """
        This class loads AutoMappers and runs them

        :param view:
        :param mapping_function:
        :param name:
        :param parameters:
        :param progress_logger:
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self.views: List[str] = []
        # noinspection Mypy
        self._setDefault(view=None)

        self.mapping_function: AutoMapperFunction = mapping_function

        # noinspection Mypy
        kwargs = self._input_kwargs
        # remove mapping_function since that is not serializable
        kwargs = {
            key: value for key, value in kwargs.items() if key != "mapping_function"
        }
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        mapping_function: AutoMapperFunction = self.getMappingFunction()

        # run the mapping function to get an AutoMapper
        incoming_parameters: Optional[Dict[str, Any]] = self.getParameters()
        parameters: Dict[str, Any] = (
            incoming_parameters.copy() if incoming_parameters else {}
        )
        parameters["view"] = view

        auto_mappers: List[AutoMapperBase] = []
        auto_mapper = mapping_function(parameters)
        if isinstance(auto_mapper, list):
            auto_mappers = auto_mappers + auto_mapper
        else:
            auto_mappers.append(auto_mapper)

        assert isinstance(auto_mappers, list)

        i: int = 0
        count: int = len(auto_mappers)
        for automapper in auto_mappers:
            i += 1
            assert isinstance(automapper, AutoMapper)
            if automapper.view:
                self.views.append(automapper.view)
            self.logger.info(
                f"---- Running AutoMapper {i} of {count} [{automapper}] ----"
            )
            table_names: List[str] = df.sql_ctx.tableNames()

            # if view exists then drop it
            if (
                automapper.view
                and automapper.use_single_select
                and not automapper.reuse_existing_view
                and automapper.view in table_names
            ):
                self.logger.info(f"Dropping view {automapper.view}")
                df.sql_ctx.dropTempTable(tableName=automapper.view)

            progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
            try:
                if progress_logger is not None:
                    progress_logger.start_mlflow_run(
                        run_name=str(automapper), is_nested=True
                    )

                automapper.transform(df=df)

                if progress_logger is not None:
                    progress_logger.end_mlflow_run()
            except Exception as e:
                error_msg = f"Error running automapper {automapper.view}"
                if progress_logger is not None:
                    progress_logger.log_exception(
                        event_name=str(automapper), event_text=error_msg, ex=e
                    )
                    progress_logger.end_mlflow_run(status=RunStatus.FAILED)
                raise FrameworkMappingRunnerException(msg=error_msg, exception=e) from e

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming
    def getViews(self) -> List[str]:
        return self.views

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMappingFunction(self) -> AutoMapperFunction:
        return self.mapping_function

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self.getView()}"



import os
import string
from pathlib import Path
import random
from shutil import rmtree
from typing import Dict, Any, Callable, Union, List, cast

import mlflow  # type: ignore
import pytest
from mlflow.entities import Run, RunStatus  # type: ignore
from pyspark.ml import Transformer
from spark_auto_mapper.automappers.automapper_base import AutoMapperBase
from spark_pipeline_framework.event_loggers.event_logger import EventLogger

from spark_pipeline_framework.transformers.framework_json_exporter.v1.framework_json_exporter import (
    FrameworkJsonExporter,
)

from spark_pipeline_framework.transformers.framework_mapping_runner.v1.framework_mapping_runner import (
    FrameworkMappingLoader,
)

from spark_pipeline_framework.proxy_generator.python_transformer_helpers import (
    get_python_function_from_location,
)

from spark_pipeline_framework.transformers.framework_if_else_transformer.v1.framework_if_else_transformer import (
    FrameworkIfElseTransformer,
)

from spark_pipeline_framework.progress_logger.progress_logger import (
    ProgressLogger,
    MlFlowConfig,
)

from library.features.carriers_python.v1.features_carriers_python_v1 import (
    FeaturesCarriersPythonV1,
)

from library.features.carriers.v1.features_carriers_v1 import FeaturesCarriersV1
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType
from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import (
    FrameworkCsvLoader,
)

from spark_pipeline_framework.pipelines.v2.framework_pipeline import FrameworkPipeline
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)

from tests.conftest import clean_spark_session


class SimplePipeline(FrameworkPipeline):
    def __init__(self, parameters: Dict[str, Any], progress_logger: ProgressLogger):
        super(SimplePipeline, self).__init__(
            parameters=parameters,
            progress_logger=progress_logger,
            run_id="12345678",
            client_name="client_foo",
            vendor_name="vendor_foo",
        )

        mapping_function: Callable[
            [Dict[str, Any]], Union[AutoMapperBase, List[AutoMapperBase]]
        ] = get_python_function_from_location(
            location=str(parameters["feature_path"]),
            import_module_name=".mapping",
            function_name="mapping",
        )

        self.transformers = self.create_steps(
            cast(
                List[Transformer],
                [
                    FrameworkCsvLoader(
                        view="flights",
                        file_path=parameters["flights_path"],
                        parameters=parameters,
                        progress_logger=progress_logger,
                    ),
                    FeaturesCarriersV1(
                        parameters=parameters, progress_logger=progress_logger
                    ),
                    FrameworkIfElseTransformer(
                        enable=True,
                        progress_logger=progress_logger,
                        stages=[
                            FeaturesCarriersPythonV1(
                                parameters=parameters, progress_logger=progress_logger
                            ),
                        ],
                    ),
                    FrameworkMappingLoader(
                        view="members",
                        mapping_function=mapping_function,
                        parameters=parameters,
                        progress_logger=progress_logger,
                    ),
                    FrameworkJsonExporter(
                        file_path=parameters["export_path"],
                        view="flights",
                        name="export flights as json",
                        parameters=parameters,
                        progress_logger=progress_logger,
                    ),
                ],
            )
        )


class MappingPipeline(FrameworkPipeline):
    def __init__(self, parameters: Dict[str, Any], progress_logger: ProgressLogger):
        super(MappingPipeline, self).__init__(
            parameters=parameters,
            progress_logger=progress_logger,
            run_id="987654321",
            client_name="client_bar",
            vendor_name="vendor_bar",
        )
        mapping_function: Callable[
            [Dict[str, Any]], Union[AutoMapperBase, List[AutoMapperBase]]
        ] = get_python_function_from_location(
            location=str(parameters["feature_path"]),
            import_module_name=".mapping",
            function_name="mapping",
        )

        self.transformers = self.create_steps(
            cast(
                List[FrameworkTransformer],
                [
                    FrameworkMappingLoader(
                        view="members",
                        mapping_function=mapping_function,
                        parameters=parameters,
                        progress_logger=progress_logger,
                    )
                ],
            )
        )


@pytest.fixture(scope="module")
def test_setup() -> None:
    data_dir = Path(__file__).parent
    temp_dir = data_dir.joinpath("temp")
    if os.path.isdir(temp_dir):
        rmtree(temp_dir)
    os.makedirs(temp_dir)


def test_progress_logger_with_mlflow(
    spark_session: SparkSession, test_setup: Any
) -> None:
    # Arrange
    clean_spark_session(spark_session)
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_dir: Path = data_dir.joinpath("temp")
    flights_path: str = f"file://{data_dir.joinpath('flights.csv')}"
    export_path: str = str(temp_dir.joinpath("ouptput").joinpath("flights.json"))

    schema = StructType([])

    spark_session.createDataFrame(spark_session.sparkContext.emptyRDD(), schema)

    spark_session.sql("DROP TABLE IF EXISTS default.flights")

    spark_session.createDataFrame(
        [
            (1, "Qureshi", "Imran"),
            (2, "Vidal", "Michael"),
        ],  # noqa: E231
        ["member_id", "last_name", "first_name"],
    ).createOrReplaceTempView("patients")

    source_df: DataFrame = spark_session.table("patients")

    df = source_df.select("member_id")
    df.createOrReplaceTempView("members")

    # Act
    parameters = {
        "flights_path": flights_path,
        "feature_path": data_dir.joinpath(
            "library/features/carriers_multiple_mappings/v1"
        ),
        "foo": "bar",
        "view2": "my_view_2",
        "export_path": export_path,
        "conn_str": "jdbc:mysql://username:Im5CYsCO923GFAebv6bf@warehouse-mysql.server:3306/"
        + "schema?rewriteBatchedStatements=true",
    }

    flow_run_name = "fluffy-fox"

    mlflow_tracking_url = temp_dir.joinpath("mlflow")
    # mlflow_tracking_url = "http://mlflow:5000"
    artifact_url = str(temp_dir.joinpath("mlflow_artifacts"))
    random_string = "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(20)
    )
    experiment_name = random_string

    mlflow_config = MlFlowConfig(
        parameters=parameters,
        experiment_name=experiment_name,
        flow_run_name=flow_run_name,
        mlflow_tracking_url=str(mlflow_tracking_url),
        artifact_url=artifact_url,
    )

    with ProgressLogger(mlflow_config=mlflow_config) as progress_logger:
        pipeline: SimplePipeline = SimplePipeline(
            parameters=parameters, progress_logger=progress_logger
        )
        transformer = pipeline.fit(df)
        transformer.transform(df)

    # assert we have an experiment created in mlflow
    experiment = mlflow.get_experiment_by_name(name=experiment_name)
    assert experiment is not None, "the mlflow experiment was not created"
    # assert the experiment has one parent run and 7 nested runs
    runs = mlflow.search_runs(
        experiment_ids=[experiment.experiment_id], output_format="list"
    )
    assert len(runs) == 9, "there should be 9 runs total, 1 parent and 8 nested"
    parent_runs = [
        run for run in runs if run.data.tags.get("mlflow.parentRunId") is None
    ]
    assert len(parent_runs) == 1
    nested_runs = [
        run for run in runs if run.data.tags.get("mlflow.parentRunId") is not None
    ]
    assert len(nested_runs) == 8
    # assert that the parent run has the params
    parent_run: Run = parent_runs[0]
    assert (
        parent_run.data.params.get("flights_path") == flights_path
    ), "parent run should have the flights_path parameter set"

    # assert that load and export runs have the data param set
    csv_loader_run = [
        run
        for run in nested_runs
        if "FrameworkCsvLoader" in run.data.tags.get("mlflow.runName")
    ]
    assert len(csv_loader_run) == 1
    assert (
        csv_loader_run[0].data.params.get("data_path") == flights_path
    ), "csv loader run should have 'data_path` param set"
    json_export_run = [
        run
        for run in nested_runs
        if "FrameworkJsonExporter" in run.data.tags.get("mlflow.runName")
    ]
    assert len(json_export_run) == 1
    assert (
        json_export_run[0].data.params.get("data_export_path") == export_path
    ), "export run should have 'data_export_path` param set"


def test_progress_logger_without_mlflow(
    spark_session: SparkSession, test_setup: Any
) -> None:
    clean_spark_session(spark_session)
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_dir: Path = data_dir.joinpath("temp")

    flights_path: str = f"file://{data_dir.joinpath('flights.csv')}"
    export_path: str = str(temp_dir.joinpath("ouptput").joinpath("flights.json"))

    schema = StructType([])

    spark_session.createDataFrame(spark_session.sparkContext.emptyRDD(), schema)

    spark_session.sql("DROP TABLE IF EXISTS default.flights")

    spark_session.createDataFrame(
        [
            (1, "Qureshi", "Imran"),
            (2, "Vidal", "Michael"),
        ],  # noqa: E231
        ["member_id", "last_name", "first_name"],
    ).createOrReplaceTempView("patients")

    source_df: DataFrame = spark_session.table("patients")

    df = source_df.select("member_id")
    df.createOrReplaceTempView("members")

    # Act
    parameters = {
        "flights_path": flights_path,
        "feature_path": data_dir.joinpath(
            "library/features/carriers_multiple_mappings/v1"
        ),
        "foo": "bar",
        "view2": "my_view_2",
        "export_path": export_path,
    }

    with ProgressLogger() as progress_logger:
        pipeline: SimplePipeline = SimplePipeline(
            parameters=parameters, progress_logger=progress_logger
        )
        transformer = pipeline.fit(df)
        transformer.transform(df)
    # semi hack -- reset the tracking url s
    mlflow.set_tracking_uri(uri="")
    experiments = mlflow.list_experiments()
    assert len(experiments) == 1
    assert experiments[0].name == "Default"

    mlflow_default_dir: Path = data_dir.joinpath("mlruns")
    if os.path.isdir(mlflow_default_dir):
        rmtree(mlflow_default_dir)


def test_progress_logger_mlflow_error_handling(test_setup: Any) -> None:
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_dir: Path = data_dir.joinpath("temp")
    event_log_path = temp_dir.joinpath("event_log")

    class FileEventLogger(EventLogger):
        def __init__(self, log_path: Path):
            self.log_path = log_path
            os.makedirs(self.log_path)

        def log_event(self, event_name: str, event_text: str) -> None:
            log_file_path: Path = self.log_path.joinpath(
                f"{event_name.replace(' ', '_')}"
            )
            with open(log_file_path, "w") as text_file:
                text_file.write(event_text)

    file_event_logger = FileEventLogger(log_path=event_log_path)

    parameters = {"foo": "bar", "view2": "my_view_2"}

    mlflow_tracking_url = temp_dir.joinpath("mlflow")
    artifact_url = str(temp_dir.joinpath("mlflow_artifacts"))
    experiment_name: str = "error_tests"

    mlflow_config = MlFlowConfig(
        parameters=parameters,
        experiment_name=experiment_name,
        flow_run_name="run",
        mlflow_tracking_url=str(mlflow_tracking_url),
        artifact_url=artifact_url,
    )
    with ProgressLogger(
        mlflow_config=mlflow_config, event_loggers=[file_event_logger]
    ) as progress_logger:
        # log a param with the same key and different values and
        # verify we get other params logged and notification of failure
        params = {"bar": "foo", "foo": "foo", "log": "this"}
        progress_logger.log_params(params=params)

        # test with different data types to see if we get an error when logging
        progress_logger.log_param("page_size", 200)  # type: ignore
        progress_logger.log_param("catalog_entry_name", None)  # type: ignore

        progress_logger.log_metric(name=200, time_diff_in_minutes=1)  # type: ignore

    # assert that there are all the params logged except for the one causing the error
    experiment = mlflow.get_experiment_by_name(name=experiment_name)
    assert experiment is not None, "the mlflow experiment was not created"

    runs: List[Run] = mlflow.search_runs(
        experiment_ids=[experiment.experiment_id], output_format="list"
    )
    assert len(runs) == 1
    run: Run = runs[0]
    assert run.info.status == RunStatus.to_string(RunStatus.FINISHED)

    # assert that the 'log' param was set properly
    log_param_value = run.data.params.get("log")
    assert log_param_value == "this"

    # assert that an event notification was sent out
    event_log_files = os.listdir(event_log_path)
    assert len(event_log_files) == 1



from typing import Any


class BaseSlackClient:
    def post_message_to_slack(
        self, text: str, blocks: Any = None, use_conversation_threads: bool = True
    ) -> Any:
        raise NotImplementedError



from typing import List, Optional, Union, Callable

from pyspark.ml import Transformer
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.transformers.framework_if_else_transformer.v1.framework_if_else_transformer import (
    FrameworkIfElseTransformer,
)


def run_if_else(
    *,
    enable: Optional[Union[bool, Callable[[DataFrame], bool]]] = None,
    enable_if_view_not_empty: Optional[
        Union[str, Callable[[Optional[str]], str]]
    ] = None,
    stages: Union[List[Transformer], Callable[[], List[Transformer]]],
    else_stages: Optional[
        Union[List[Transformer], Callable[[], List[Transformer]]]
    ] = None,
) -> FrameworkIfElseTransformer:
    """
    If enable flag is true then runs stages else runs else_stages
    :param enable: a boolean or a function that takes a DataFrame and returns a boolean
    :param stages: list of transformers or a function that returns a list of transformers
    :param else_stages: list of transformers or a function that returns a list of transformers
    :param enable_if_view_not_empty: enable if this view is not empty
    """
    return FrameworkIfElseTransformer(
        enable=enable,
        enable_if_view_not_empty=enable_if_view_not_empty,
        stages=stages,
        else_stages=else_stages,
    )



import os

from smart_open import open as smart_open  # type: ignore
from typing import Optional, Dict, Any, List

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger

from spark_pipeline_framework.logger.yarn_logger import get_logger

from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.file_helpers import isfile, listdir

pipeline_validation_df_name = "pipeline_validation"


class FrameworkValidationTransformer(FrameworkTransformer):
    """
    The FrameworkValidationTransformer will run all SQL based validations in the paths specified and optionally fail the
    pipeline if validations fail. The results of validation will be stored in a dataframe and persisted with a temp table
    named 'pipeline_validation'
    """

    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        validation_source_path: str,
        validation_queries: List[str],
        fail_on_validation: bool = False,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> None:
        """

        :param validation_source_path: the root path to the validation queries
        :param validation_queries: a list of validation queries paths relative to the root path, can be the path of a file relative to the root path, or a path to a directory relative to root path
        :param fail_on_validation: if True, fails the pipeline at this transformer, defaults to False
        :param name: a name for the transformer step
        :param parameters: parameters
        :param progress_logger: the logger to use for logging
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        self.logger = get_logger(__name__)

        if not validation_source_path:
            raise ValueError("validation source path is None or empty")

        if not validation_queries:
            raise ValueError("validation_queries is None")

        self.validation_source_path: Param[str] = Param(
            self, "validation_source_path", ""
        )
        self._setDefault(validation_source_path=None)

        self.validation_queries: Param[List[str]] = Param(
            self, "validation_queries", ""
        )
        self._setDefault(validation_queries=None)

        self.fail_on_validation: Param[bool] = Param(self, "fail_on_validation", "")
        self._setDefault(fail_on_validation=False)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        validation_source_path = self.getValidationSourcePath()
        validation_queries = self.getValidationQueries()
        fail_on_validation = self.getFailOnValidation()

        for query in validation_queries:
            path = os.path.join(validation_source_path, query)
            self._validate(path, df)

        if fail_on_validation:
            errors_df = df.sql_ctx.sql(
                f"SELECT * from {pipeline_validation_df_name} where is_failed == 1"
            )
            error_count = errors_df.count()
            assert (
                error_count == 0
            ), f"Pipeline failed validation, there were {error_count} errors. See additional logs or details"
        return df

    def _validate(
        self,
        path: str,
        df: DataFrame,
    ) -> None:
        validation_df = self.get_validation_df(df)
        if isfile(path):
            self.logger.info(f"Path: {path} is a file")
            with smart_open(path, "r") as query_file:
                self.logger.info(f"Executing validation query: {path}")
                query_text = query_file.read()
                query_text = query_text.upper().replace(
                    "SELECT", f"SELECT '{path}' as query,\n"
                )
                if validation_df:
                    validation_df = validation_df.union(df.sql_ctx.sql(query_text))
                    validation_df.createOrReplaceTempView(pipeline_validation_df_name)
                else:
                    validation_df = df.sql_ctx.sql(query_text)
                    validation_df.createOrReplaceTempView(pipeline_validation_df_name)
        else:
            self.logger.info(f"Path: {path} is a directory, getting paths")
            paths = listdir(path)
            self.logger.info(f"Got paths: {paths}")
            for child_path in paths:
                self._validate(child_path, df)

    # noinspection PyMethodMayBeStatic
    def get_validation_df(self, df: DataFrame) -> Optional[DataFrame]:
        validation_df: Optional[DataFrame] = None
        tables_df = df.sql_ctx.tables().filter(
            f"tableName ='{pipeline_validation_df_name}'"
        )
        if tables_df.count() == 1:
            validation_df = df.sql_ctx.table(pipeline_validation_df_name)
        return validation_df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getValidationSourcePath(self) -> str:
        return self.getOrDefault(self.validation_source_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getValidationQueries(self) -> List[str]:
        return self.getOrDefault(self.validation_queries)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFailOnValidation(self) -> bool:
        return self.getOrDefault(self.fail_on_validation)



import collections
import concurrent.futures
import threading
import traceback
from typing import List, AsyncIterator, Tuple, Any, Dict, Optional, OrderedDict

from bounded_pool_executor import BoundedThreadPoolExecutor
from pyspark.ml import Pipeline, Transformer
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger


class ParallelPipelineExecutor:
    """
    Runs SparkML Pipelines in parallel
    """

    def __init__(
        self,
        progress_logger: Optional[ProgressLogger],
        max_tasks: int = 3,
        parallel_mode: bool = False,
        parameters: Optional[Dict[str, Any]] = None,
    ):
        """

        :param progress_logger: optional parameter which is used for logging progress
        :param max_tasks: maximum tasks to run at one time
        """
        super().__init__()
        self.logger = get_logger(__name__)
        self.dictionary: OrderedDict[str, List[Transformer]] = collections.OrderedDict()
        self.max_tasks: int = max_tasks
        self.progress_logger: Optional[ProgressLogger] = progress_logger
        self.parallel_mode: bool = parallel_mode
        self.parameters = parameters

    def append(self, name: str, list_of_stages: List[Transformer]) -> None:
        """
        appends a list of stages to dictionary
        :param name: name to associate with this list
        :param list_of_stages:
        """
        self.logger.info(f" dictionary {self.dictionary}")
        self.logger.info(f"Adding name={name}, count={len(self.dictionary)}")
        if name in self.dictionary:
            raise ValueError(f"There is already an entry for ({name}).")
        self.dictionary[name] = list_of_stages

    async def transform_async(
        self, df: DataFrame, spark_session: SparkSession
    ) -> AsyncIterator[Tuple[str, DataFrame]]:
        """
        runs the pipelines asynchronously
        limits number of tasks running at one time to max_tasks
        :param spark_session:
        :param df: dataframe to use for transforming
        :return: an async iterator that you can use to retrieve the results
                    while the rest of the transformers are still transforming
        """
        number_of_concurrent_tasks = min(self.max_tasks, len(self.dictionary))

        items = [(name, stages) for name, stages in self.dictionary.items()]

        # https://docs.python.org/3.7/library/concurrent.futures.html
        with BoundedThreadPoolExecutor(
            max_workers=number_of_concurrent_tasks
        ) as executor:
            track_futures = {
                executor.submit(
                    self._run_pipeline_transform_async,
                    name=item[0],
                    stages=item[1],
                    df=df,
                    spark_session=spark_session,
                    progress_logger=self.progress_logger,
                )
                for item in items
            }
            for future_item in concurrent.futures.as_completed(track_futures):
                try:
                    name, df = future_item.result()
                    yield name, df
                except Exception as e:
                    self.logger.info("ERROR transform_async {0}".format(str(e)))
                    err = traceback.format_exc()
                    self.logger.info("TRACEBACK: {0}".format(err))
                    raise

    def _run_pipeline_transform_async(
        self,
        name: str,
        stages: List[Transformer],
        df: DataFrame,
        spark_session: SparkSession,
        progress_logger: ProgressLogger,
    ) -> Tuple[str, DataFrame]:
        """
        runs transform on the pipeline
        :param spark_session:
        :param name: name of pipeline
        :param stages: stages to run
        :param df: dataframe to use
        :return: name, dataframe
        """
        try:
            self.logger.info(f"Started transforming {name}")
            result_df: DataFrame = self._run_transform(
                name, df, stages, spark_session, progress_logger, self.parallel_mode
            )

            self.logger.info(f"Finished transforming {name}")
            return name, result_df
        except Exception as e:
            self.logger.info("ERROR transforming for {0}: {1}".format(name, str(e)))
            err = traceback.format_exc()
            self.logger.info("TRACEBACK: {0}".format(err))
            raise

    @staticmethod
    def _run_transform(
        name: str,
        df: DataFrame,
        stages: List[Transformer],
        spark_session: SparkSession,
        progress_logger: ProgressLogger,
        parallel_mode: bool,
    ) -> DataFrame:
        logger = get_logger(__name__)
        logger.info(f"{name} in running in thread id={threading.get_ident()}")
        spark_session.sparkContext.setJobDescription(f"fit and transform {name}")
        with ProgressLogMetric(name=name, progress_logger=progress_logger):
            if parallel_mode:
                result_df = ParallelPipelineExecutor._handle_transform_mode(df, stages)
            else:
                result_df = Pipeline(stages=stages).fit(df).transform(df)
            return result_df

    @staticmethod
    def _handle_transform_mode(df: DataFrame, stages: List[Transformer]) -> DataFrame:
        new_df = df
        model = Pipeline(stages=stages).fit(new_df)
        result_df = model.transform(new_df)
        return result_df



from typing import Dict

from spark_auto_mapper.type_definitions.defined_types import (
    AutoMapperTextInputType,
    AutoMapperAnyDataType,
)

from library.translators.my_translator.translators_my_translator import (
    TranslatorsMyTranslator,
)


def test_translator() -> None:
    my_dict: Dict[
        AutoMapperTextInputType, AutoMapperAnyDataType
    ] = TranslatorsMyTranslator().get_mapping()
    print(my_dict)

    assert my_dict == {"foo1": "bar1", "foo2": "bar2"}



from pathlib import Path
from typing import Any, Callable, Dict, Optional, Union

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql import DataFrameReader
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.streaming import DataStreamReader
from pyspark.sql.utils import AnalysisException

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.file_modes import FileReadModes


class FrameworkParquetLoader(FrameworkTransformer):
    """
    Loads a view from parquet
    """

    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        merge_schema: bool = False,
        limit: Optional[int] = None,
        mode: str = FileReadModes.MODE_PERMISSIVE,
        stream: bool = False,
        delta_lake_table: Optional[str] = None,
    ):
        """
        Loads given view from parquet


        :param view: view to load from parquet
        :param file_path: where to load from
        :param name: a name for the transformer step
        :param parameters: parameters
        :param progress_logger: the logger to use for logging
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        assert mode in FileReadModes.MODE_CHOICES

        assert (
            isinstance(file_path, Path)
            or isinstance(file_path, str)
            or callable(file_path)
        ), type(file_path)
        assert file_path

        self.logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.file_path: Param[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.merge_schema: Param[bool] = Param(self, "merge_schema", "")
        self._setDefault(merge_schema=None)

        self.limit: Param[Optional[int]] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        self.stream: Param[bool] = Param(self, "stream", "")
        self._setDefault(stream=False)

        self.delta_lake_table: Param[Optional[str]] = Param(
            self, "delta_lake_table", ""
        )
        self._setDefault(delta_lake_table=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        file_path: Union[
            Path, str, Callable[[Optional[str]], Union[Path, str]]
        ] = self.getFilePath()
        if callable(file_path):
            file_path = file_path(self.loop_id)
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        merge_schema: bool = self.getMergeSchema()
        limit: Optional[int] = self.getLimit()
        stream: bool = self.getStream()

        if progress_logger:
            progress_logger.write_to_log(
                f"Loading parquet file for view {view}: {file_path}"
            )
            progress_logger.log_param(key="data_path", value=str(file_path))

        with ProgressLogMetric(
            name=f"{name or view}_table_loader", progress_logger=progress_logger
        ):
            try:
                df_reader: Union[DataFrameReader, DataStreamReader] = (
                    df.sql_ctx.read if not stream else df.sql_ctx.readStream
                )

                df_reader = df_reader.option("mode", self.getMode())

                if merge_schema is True:
                    final_df = (
                        df_reader.option("mergeSchema", "true")
                        .format("parquet")
                        .load(path=str(file_path))
                    )
                else:
                    final_df = df_reader.format("parquet").load(path=str(file_path))

                assert (
                    "_corrupt_record" not in final_df.columns
                ), f"Found _corrupt_record after reading the file: {file_path}. "

                if limit is not None and limit >= 0:
                    final_df = final_df.limit(limit)

                # store new data frame in the view
                final_df.createOrReplaceTempView(view)
            except AnalysisException as e:
                self.logger.error(
                    f"File load failed. Location: {file_path} may be empty"
                )
                raise e
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMergeSchema(self) -> bool:
        return self.getOrDefault(self.merge_schema)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> Optional[int]:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMode(self) -> str:
        return self.getOrDefault(self.mode)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getStream(self) -> bool:
        return self.getOrDefault(self.stream)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDeltaLakeTable(self) -> Optional[str]:
        return self.getOrDefault(self.delta_lake_table)



import re
from logging import Logger
from pathlib import Path
from typing import Union, List, Optional, Dict, Any, Callable

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql import DataFrame
from pyspark.sql.functions import input_file_name
from pyspark.sql.readwriter import DataFrameReader
from pyspark.sql.streaming import DataStreamReader
from pyspark.sql.types import StructType

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.file_modes import FileReadModes


class FrameworkLocalFileLoader(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        file_path: Union[
            str, List[str], Path, Callable[[Optional[str]], Union[Path, str]]
        ],
        delimiter: str = ",",
        limit: Optional[int] = None,
        has_header: bool = True,
        infer_schema: bool = False,
        cache_table: bool = True,
        schema: Optional[StructType] = None,
        clean_column_names: bool = False,
        create_file_path: bool = False,
        use_schema_from_view: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        mode: str = FileReadModes.MODE_PERMISSIVE,
        stream: bool = False,
        delta_lake_table: Optional[str] = None,
    ) -> None:
        """
        Base class that handles loading of files


        :param view: view to populate with the data in the file
        :param file_path: where to store the FHIR resources retrieved.  Can be either a path or a function
                    that returns path. Function is passed a loop number if inside a loop or else None
        :param delimiter:
        :param limit:
        :param has_header:
        :param infer_schema:
        :param cache_table: whether to cache the table in memory after loading it
        :param schema: schema to use when loading the data
        :param clean_column_names: whether to remove invalid characters from column names
                                    (i.e., characters that Spark does not like)
        :param create_file_path: whether to create the file path if it does not exist
        :param use_schema_from_view: use schema from this view when loading data
        :param mode: file loading mode
        :param stream: whether to stream the data
        :param delta_lake_table: store data in this delta lake
        """
        super().__init__(
            name=name,
            parameters=parameters,
            progress_logger=progress_logger,
        )

        assert mode in FileReadModes.MODE_CHOICES

        self.logger: Logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=None)

        self.file_path: Param[
            Union[str, List[str], Path, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.schema: Param[StructType] = Param(self, "schema", "")
        self._setDefault(schema=None)

        self.clean_column_names: Param[bool] = Param(self, "clean_column_names", "")
        self._setDefault(clean_column_names=False)

        self.cache_table: Param[bool] = Param(self, "cache_table", "")
        self._setDefault(cache_table=True)

        self.limit: Param[Optional[int]] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.infer_schema: Param[bool] = Param(self, "infer_schema", "")
        self._setDefault(infer_schema=False)

        self.create_file_path: Param[bool] = Param(self, "create_file_path", "")
        self._setDefault(create_file_path=False)

        self.use_schema_from_view: Param[str] = Param(self, "use_schema_from_view", "")
        self._setDefault(use_schema_from_view=None)

        self.delimiter: Param[str] = Param(self, "delimiter", "")
        self._setDefault(delimiter=",")

        self.has_header: Param[bool] = Param(self, "has_header", "")
        self._setDefault(has_header=True)

        self.stream: Param[bool] = Param(self, "stream", "")
        self._setDefault(stream=False)

        self.delta_lake_table: Param[Optional[str]] = Param(
            self, "delta_lake_table", ""
        )
        self._setDefault(delta_lake_table=None)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        if not file_path:
            raise ValueError("file_path is None or empty")

        self.logger.info(f"Received file_path: {file_path}")

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view = self.getView()
        file_path: Union[
            str, List[str], Path, Callable[[Optional[str]], Union[Path, str]]
        ] = self.getFilepath()
        schema = self.getSchema()
        clean_column_names = self.getCleanColumnNames()
        cache_table = self.getCacheTable()
        infer_schema = self.getInferSchema()
        limit: Optional[int] = self.getLimit()
        create_file_path = self.getCreateFilePath()
        use_schema_from_view = self.getUseSchemaFromView()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        stream: bool = self.getStream()

        if not file_path:
            raise ValueError(f"file_path is empty: {file_path}")

        if isinstance(file_path, str):
            if file_path.__contains__(":"):
                absolute_paths: List[str] = [file_path]
            else:
                data_dir = Path(__file__).parent.parent.joinpath("./")
                absolute_paths = [f"file://{data_dir.joinpath(file_path)}"]
        elif isinstance(file_path, Path):
            data_dir = Path(__file__).parent.parent.joinpath("./")
            absolute_paths = [f"file://{data_dir.joinpath(file_path)}"]
        elif isinstance(file_path, (list, tuple)):
            data_dir = Path(__file__).parent.parent.joinpath("./")
            absolute_paths = []
            for path in file_path:
                if ":" in path:
                    absolute_paths.append(path)
                else:
                    absolute_paths.append(f"file://{data_dir.joinpath(path)}")
        elif callable(file_path):
            absolute_paths = [str(file_path(self.loop_id))]
        else:
            raise TypeError(
                f"Unknown type '{type(file_path)}' for file_path {file_path}"
            )

        if progress_logger:
            progress_logger.write_to_log(
                f"Loading {self.getReaderFormat()} file for view {view}: {absolute_paths}, infer_schema: {infer_schema}"
            )
            progress_logger.log_param("data_path", ", ".join(absolute_paths))

        self.preprocess(df=df, absolute_paths=absolute_paths)

        df_reader: Union[DataFrameReader, DataStreamReader] = (
            df.sql_ctx.read if not stream else df.sql_ctx.readStream
        )
        df_reader = df_reader.option("mode", self.getMode())

        if schema:
            df_reader = df_reader.schema(schema)
        elif use_schema_from_view:
            df_source_view_for_schema = df.sql_ctx.table(use_schema_from_view)
            df_reader = df_reader.schema(df_source_view_for_schema.schema)
        elif infer_schema:
            df_reader = df_reader.option("inferSchema", "true")
        else:
            df_reader = df_reader.option("inferSchema", "false")

        df2: DataFrame
        df_reader = df_reader.format(self.getReaderFormat())

        for k, v in self.getReaderOptions().items():
            df_reader = df_reader.option(k, v)

        df2 = (
            df_reader.load(absolute_paths)
            if isinstance(df_reader, DataFrameReader)
            else df_reader.load(absolute_paths[0])
        )

        if create_file_path:
            df2 = df2.withColumn("file_path", input_file_name())

        if clean_column_names:
            for c in df2.columns:
                df2 = df2.withColumnRenamed(c, re.sub(r"[ ,;{}()\n\t=]", "_", c))

        assert (
            "_corrupt_record" not in df2.columns
        ), f"Found _corrupt_record after reading the file: {','.join(absolute_paths)}. "

        if limit and limit > -1:
            df2 = df2.limit(limit)

        # now save the data frame into view
        df2.createOrReplaceTempView(view)

        if cache_table:
            self.logger.info(f"Caching table {view}")
            df.sql_ctx.sql(f"CACHE TABLE {view}")

        progress_logger and progress_logger.write_to_log(
            f"Finished Loading {self.getReaderFormat()} file for View[{view}]: {absolute_paths}, "
            + f"infer_schema: {infer_schema}"
        )

        return df

    def preprocess(self, df: DataFrame, absolute_paths: List[str]) -> None:
        """
        Sub-classes can over-ride to do any pre-processing behavior

        :param df: Data Frame
        :param absolute_paths: list of paths to read from
        """

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilepath(
        self,
    ) -> Union[str, List[str], Path, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSchema(self) -> StructType:
        return self.getOrDefault(self.schema)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getCleanColumnNames(self) -> bool:
        return self.getOrDefault(self.clean_column_names)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getCacheTable(self) -> bool:
        return self.getOrDefault(self.cache_table)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getInferSchema(self) -> bool:
        return self.getOrDefault(self.infer_schema)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> Optional[int]:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getCreateFilePath(self) -> bool:
        return self.getOrDefault(self.create_file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getUseSchemaFromView(self) -> str:
        return self.getOrDefault(self.use_schema_from_view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> str:
        return self.getView()

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderFormat(self) -> str:
        raise NotImplementedError("Must implement in baseclass")

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderOptions(self) -> Dict[str, Any]:
        raise NotImplementedError("Must implement in baseclass")

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMode(self) -> str:
        return self.getOrDefault(self.mode)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getStream(self) -> bool:
        return self.getOrDefault(self.stream)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDeltaLakeTable(self) -> Optional[str]:
        return self.getOrDefault(self.delta_lake_table)



from typing import Dict, Any, Optional, Union, List, Callable

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml import Transformer
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkIfElseTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        enable: Optional[Union[bool, Callable[[DataFrame], bool]]] = None,
        enable_if_view_not_empty: Optional[
            Union[str, Callable[[Optional[str]], str]]
        ] = None,
        stages: Union[List[Transformer], Callable[[], List[Transformer]]],
        else_stages: Optional[
            Union[List[Transformer], Callable[[], List[Transformer]]]
        ] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        If enable flag is true then runs stages else runs else_stages
        :param enable: a boolean or a function that takes a DataFrame and returns a boolean
        :param stages: list of transformers or a function that returns a list of transformers
        :param else_stages: list of transformers or a function that returns a list of transformers
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.enable: Optional[Union[bool, Callable[[DataFrame], bool]]] = enable

        self.enable_if_view_not_empty: Optional[
            Union[str, Callable[[Optional[str]], str]]
        ] = enable_if_view_not_empty

        self.stages: Union[List[Transformer], Callable[[], List[Transformer]]] = stages
        self.else_stages: Optional[
            Union[List[Transformer], Callable[[], List[Transformer]]]
        ] = else_stages

        self.loop_id: Optional[str] = None

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        enable = self.enable(df) if callable(self.enable) else self.enable
        view_enable_if_view_not_empty = (
            self.enable_if_view_not_empty(self.loop_id)
            if callable(self.enable_if_view_not_empty)
            else self.enable_if_view_not_empty
        )
        enable_if_view_not_empty = (
            not df.sparkSession.table(view_enable_if_view_not_empty).isEmpty()
            if view_enable_if_view_not_empty
            else True
        )
        if (enable or enable is None) and enable_if_view_not_empty:
            stages: List[Transformer] = (
                self.stages if isinstance(self.stages, list) else self.stages()
            )
        else:
            if progress_logger is not None:
                progress_logger.write_to_log(
                    self.getName() or "FrameworkIfElseTransformer",
                    f"Skipping stages because enable {self.enable} did not evaluate to True",
                )
            stages = (
                []
                if self.else_stages is None
                else self.else_stages
                if isinstance(self.else_stages, list)
                else self.else_stages()
            )
        for stage in stages:
            if progress_logger is not None:
                progress_logger.start_mlflow_run(run_name=str(stage), is_nested=True)
            if hasattr(stage, "set_loop_id"):
                stage.set_loop_id(self.loop_id)
            df = stage.transform(df)
            if progress_logger is not None:
                progress_logger.end_mlflow_run()
        return df

    def as_dict(self) -> Dict[str, Any]:
        return {
            **(super().as_dict()),
            "enable": self.enable,
            "enable_if_view_not_empty": self.enable_if_view_not_empty,
            "stages": [s.as_dict() for s in self.stages]
            if not callable(self.stages)
            else str(self.stages),
            "else_stages": [s.as_dict() for s in self.else_stages]
            if self.else_stages and not callable(self.else_stages)
            else str(self.else_stages),
        }



from typing import List, Optional, Any, Dict, Union

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


def get_dtype(df: DataFrame, colnames: List[str]) -> Dict[str, str]:
    return dict(df.select(colnames).dtypes)


class FrameworkFillNaTransformer(FrameworkTransformer):
    """
    Replace NA/Null values with a specified value for each column in the dictionary
    """

    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        column_mapping: Dict[str, Any],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__()
        self.logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.column_mapping: Param[Dict[str, Any]] = Param(self, "column_mapping", "")
        self._setDefault(column_mapping=column_mapping)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        column_mapping: Dict[str, Any] = self.getColumnMapping()
        view: str = self.getView()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        with ProgressLogMetric(name=f"{view}_fill_na", progress_logger=progress_logger):
            self.logger.info(
                f"filling rows if any null values with replacement_value found for columns: {list(column_mapping.keys())}"
            )
            df_with_na: DataFrame = df.sql_ctx.table(view)
            df_with_filled_na = df_with_na
            data_types = get_dtype(df_with_na, list(column_mapping.keys()))

            value: Union[bool, int, float, str]
            for col, value in column_mapping.items():

                if data_types[col] != "string":
                    try:
                        value = float(value)
                    except Exception as e:
                        print(str(e))
                        print(
                            f"The data type of column: {col} is {data_types[col]}. Either cast the column as a StringType or change the type of the value you are feeding as the replacement value to a string type."
                        )

                df_with_filled_na = df_with_filled_na.na.fill(value=value, subset=[col])

            df_with_filled_na.createOrReplaceTempView(view)
        return df_with_filled_na

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getColumnMapping(self) -> Dict[str, Any]:
        return self.getOrDefault(self.column_mapping)



from os import path, makedirs, listdir
from pathlib import Path
from shutil import rmtree, copyfile
from typing import Callable, List, Optional, Set

from spark_pipeline_framework.proxy_generator.proxy_generator import ProxyGenerator


def test_can_generate_proxies() -> None:
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)

    makedirs(temp_folder)

    src_library_folder: str = path.join(data_dir, "library")
    temp_library_folder: str = path.join(data_dir, "temp", "library")

    if not path.exists(temp_library_folder):
        makedirs(temp_library_folder)

    # copytree(src=src_library_folder, dst=temp_library_folder)
    recursive_overwrite(src=src_library_folder, dst=temp_library_folder)

    # Act
    ProxyGenerator.generate_proxies(folder=temp_library_folder)

    # Assert
    data_source_path = path.join(
        temp_library_folder,
        "data_sources",
        "my_data_source",
        "v1",
        "data_sources_my_data_source_v1.py",
    )
    assert path.exists(data_source_path)

    assert path.exists
    (
        path.join(
            temp_library_folder,
            "features",
            "my_sql_feature",
            "v1",
            "features_my_sql_feature_v1.py",
        )
    )
    assert path.exists(
        path.join(
            temp_library_folder,
            "features",
            "my_python_feature",
            "v1",
            "features_my_python_feature_v1.py",
        )
    )

    assert path.exists(
        path.join(
            temp_library_folder,
            "features",
            "member_claims",
            "v1",
            "features_member_claims_v1.py",
        )
    )

    assert not path.exists(
        path.join(
            temp_library_folder,
            "features",
            "member_claims",
            "v1",
            "test",
            "features_member_claims_v1_test.py",
        )
    )


def recursive_overwrite(
    src: str, dst: str, ignore: Optional[Callable[[str, List[str]], Set[str]]] = None
) -> None:
    if path.isdir(src):
        print(f"dir={src}")
        if not path.isdir(dst):
            makedirs(dst)
        files = listdir(src)
        if ignore is not None:
            ignored = ignore(src, files)
        else:
            ignored = set()
        for f in files:
            print(f"file={f}")
            if f not in ignored:
                recursive_overwrite(path.join(src, f), path.join(dst, f), ignore)
    else:
        print(f"src={src}, dst={dst}")
        copyfile(src, dst)



from typing import Iterable, List, Any

from pyspark.sql import DataFrame


def get_pretty_table(iterable: Iterable[List[Any]], header: Iterable[str]) -> str:
    """
    Pretty prints an iterable with headers


    :param iterable: iterable to print
    :param header: headers to use
    :return:
    """
    max_len: List[int] = [len(x) for x in header]
    row: List[Any]
    for row in iterable:
        row = [row] if type(row) not in (list, tuple) else row
        for index, col in enumerate(row):
            if max_len[index] < len(str(col)):
                max_len[index] = len(str(col))
    output: str = "-" * (sum(max_len) + 1) + "\n"
    output += (
        "|"
        + "".join([h + " " * (l - len(h)) + "|" for h, l in zip(header, max_len)])
        + "\n"
    )
    output += "-" * (sum(max_len) + 1) + "\n"
    for row in iterable:
        row = [row] if type(row) not in (list, tuple) else row
        output += (
            "|"
            + "".join(
                [str(c) + " " * (l - len(str(c))) + "|" for c, l in zip(row, max_len)]
            )
            + "\n"
        )
    output += "-" * (sum(max_len) + 1) + "\n"
    return output


def get_pretty_data_frame(df: DataFrame, limit: int) -> str:
    """
    Returns the dataframe as a string

    :param df:
    :param limit:
    :return:
    """
    rows: List[List[Any]] = [list(row) for row in df.limit(limit).collect()]

    return get_pretty_table(rows, df.columns)



from pathlib import Path

from pyspark.sql import SparkSession, DataFrame
from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import (
    FrameworkCsvLoader,
)
from spark_pipeline_framework.transformers.framework_drop_rows_with_null_transformer.v1.framework_drop_rows_with_null_transformer import (
    FrameworkDropRowsWithNullTransformer,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_framework_drop_rows_with_null_transformer(spark_session: SparkSession) -> None:
    # create a dataframe with the test data
    data_dir: Path = Path(__file__).parent.joinpath("./")

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    view: str = "primary_care_protocol"
    FrameworkCsvLoader(
        view=view,
        file_path=data_dir.joinpath("primary_care_protocol.csv"),
        clean_column_names=False,
    ).transform(df)

    # ensure we have all the rows even the ones we want to drop
    result_df: DataFrame = spark_session.table(view)
    assert 7 == result_df.count()

    # drop the rows with null NPI or null Last Name
    FrameworkDropRowsWithNullTransformer(
        columns_to_check=["NPI", "Last Name"], view=view
    ).transform(df)

    # assert we get only the rows with a populated NPI
    result_df = spark_session.table(view)
    assert 1 == result_df.count()

    # ensure that no rows are dropped when there are no null values
    FrameworkDropRowsWithNullTransformer(
        columns_to_check=["NPI", "Last Name"], view=view
    ).transform(result_df)
    assert 1 == result_df.count()



from typing import Any, Dict, Optional

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_base_exporter.v1.framework_base_exporter import (
    FrameworkBaseExporter,
)
from spark_pipeline_framework.utilities.file_modes import FileWriteModes


class FrameworkJdbcExporter(FrameworkBaseExporter):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        jdbc_url: str,
        table: str,
        driver: str,
        view: Optional[str] = None,
        name: Optional[str] = None,
        mode: str = FileWriteModes.MODE_ERROR,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        limit: int = -1,
        options: Dict[str, Any] = {},
    ):
        super().__init__(
            view=view,
            name=name,
            mode=mode,
            parameters=parameters,
            progress_logger=progress_logger,
            limit=limit,
        )
        assert jdbc_url
        assert table
        assert driver

        self.logger = get_logger(__name__)

        self.options: Param[Dict[str, Any]] = Param(self, "options", "")
        self._setDefault(options=options)

        self.jdbc_url: Param[str] = Param(self, "jdbc_url", "")
        self._setDefault(jdbc_url=jdbc_url)

        self.table: Param[str] = Param(self, "table", "")
        self._setDefault(table=table)

        self.driver: Param[str] = Param(self, "driver", "")
        self._setDefault(driver=driver)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getJdbcUrl(self) -> str:
        return self.getOrDefault(self.jdbc_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getTable(self) -> str:
        return self.getOrDefault(self.table)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDriver(self) -> str:
        return self.getOrDefault(self.driver)

    def getFormat(self) -> str:
        return "jdbc"

    def getOptions(self) -> Dict[str, Any]:
        options: Dict[str, Any] = self.getOrDefault(self.options).copy()
        parameter_options: Dict[str, Any] = {
            "url": self.getJdbcUrl(),
            "dbtable": self.getTable(),
            "driver": self.getDriver(),
        }
        options.update(parameter_options)
        return options



import json
from typing import Any, Dict, List, Union, Optional

from pyspark.ml.base import Transformer
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.FriendlySparkException import (
    FriendlySparkException,
)
from spark_pipeline_framework.utilities.class_helpers import ClassHelpers
from spark_pipeline_framework.utilities.pipeline_helper import create_steps


class FrameworkPipeline(Transformer):
    def __init__(
        self, parameters: Dict[str, Any], progress_logger: ProgressLogger
    ) -> None:
        """
        Base class for all pipelines
        :param parameters:
        :param progress_logger:
        """
        super(FrameworkPipeline, self).__init__()
        self.transformers: List[Transformer] = []
        self.steps: List[Union[Transformer, List[Transformer]]] = []
        self.__parameters: Dict[str, Any] = parameters
        self.progress_logger: ProgressLogger = progress_logger
        self.loop_id: Optional[str] = None

    @property
    def parameters(self) -> Dict[str, Any]:
        return self.__parameters

    # noinspection PyUnusedLocal
    def fit(self, df: DataFrame) -> "FrameworkPipeline":
        return self

    def _transform(self, df: DataFrame) -> DataFrame:
        # if steps are defined but not transformers then convert steps to transformers first
        if len(self.steps) > 0 and len(self.transformers) == 0:
            self.transformers = self.create_steps(self.steps)
        # get the logger to use
        logger = get_logger(__name__)
        count_of_transformers: int = len(self.transformers)
        i: int = 0
        pipeline_name: str = self.__class__.__name__
        self.progress_logger.log_event(
            event_name=pipeline_name, event_text=f"Starting Pipeline {pipeline_name}"
        )
        for transformer in self.transformers:
            assert isinstance(transformer, Transformer), type(transformer)
            # assert isinstance(transformer, FrameworkTransformer), type(transformer)
            try:
                i += 1
                logger.info(
                    f"---- Running pipeline [{pipeline_name}] transformer [{transformer}]  "
                    f"({i} of {count_of_transformers}) ----"
                )
                if hasattr(transformer, "set_loop_id"):
                    transformer.set_loop_id(self.loop_id)

                with ProgressLogMetric(
                    progress_logger=self.progress_logger,
                    name=str(transformer) or "unknown",
                ):
                    self.progress_logger.log_event(
                        pipeline_name, event_text=f"Running pipeline step {transformer}"
                    )
                    df = transformer.transform(dataset=df)
            except Exception as e:
                if hasattr(transformer, "getName"):
                    # noinspection Mypy
                    stage_name = transformer.getName()
                else:
                    stage_name = transformer.__class__.__name__
                logger.error(
                    f"!!!!!!!!!!!!! pipeline [{pipeline_name}] transformer [{stage_name}] threw exception !!!!!!!!!!!!!"
                )
                # use exception chaining to add stage name but keep original exception
                friendly_spark_exception: FriendlySparkException = (
                    FriendlySparkException(exception=e, stage_name=stage_name)
                )
                error_messages: List[str] = (
                    friendly_spark_exception.message.split("\n")
                    if friendly_spark_exception.message
                    else []
                )
                for error_message in error_messages:
                    logger.error(msg=error_message)

                if hasattr(transformer, "getSql"):
                    # noinspection Mypy
                    logger.error(transformer.getSql())
                logger.error(
                    "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
                )
                self.progress_logger.log_exception(
                    event_name=pipeline_name,
                    event_text=f"Exception in Stage={stage_name}",
                    ex=e,
                )
                raise friendly_spark_exception from e
        self.progress_logger.log_event(
            event_name=pipeline_name, event_text=f"Finished Pipeline {pipeline_name}"
        )
        return df

    # noinspection PyMethodMayBeStatic
    def create_steps(
        self,
        my_list: Union[
            List[Transformer],
            List[FrameworkTransformer],
            List[Union[Transformer, List[Transformer]]],
            List[Union[FrameworkTransformer, List[FrameworkTransformer]]],
            # List[DefaultParamsReadable[Any]],
        ],
    ) -> List[Transformer]:
        return create_steps(my_list)

    def finalize(self) -> None:
        pass

    def as_dict(self) -> Dict[str, Any]:
        return {
            "short_type": self.__class__.__name__,
            "type": ClassHelpers.get_full_name_of_instance(self),
            # self.parameters is a subclass of dict so json.dumps thinks it can't serialize it
            "params": {
                k: v if not hasattr(v, "as_dict") else v.as_dict()
                for k, v in self.parameters.items()
            },
            "steps": [
                s.as_dict() if not isinstance(s, list) else [s1.as_dict() for s1 in s]
                for s in self.steps
            ],
        }

    def __str__(self) -> str:
        return json.dumps(self.as_dict(), default=str)

    def set_loop_id(self, loop_id: str) -> None:
        """
        Set when running inside a FrameworkLoopTransformer

        :param loop_id: loop id
        """
        self.loop_id = loop_id



from logging import Logger
from pathlib import Path
from typing import Union, List, Optional, Dict, Any, NamedTuple, Callable

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql import DataFrameReader
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col, trim
from pyspark.sql.types import DataType, Row

from spark_pipeline_framework.logger.yarn_logger import get_logger

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger

from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.file_helpers import get_absolute_paths


class ColumnSpec(NamedTuple):
    """
    The definition of a column for fixed width file formats
        column_name: the name of the column

        start_pos: the starting position of the data for this column in the file

        length: the length of the data for this column

        data_type: the data type for the data in the column, e.g. StringType(), IntegerType()

    example:
      ColumnSpec(column_name="id", start_pos=1, length=3, data_type=StringType())
    """

    column_name: str
    start_pos: int
    length: int
    data_type: DataType


class FrameworkFixedWidthLoader(FrameworkTransformer):
    """
    Load fixed width files into a dataframe by specifying the path to the file and the schema
    """

    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        file_path: Union[
            str, List[str], Path, Callable[[Optional[str]], Union[Path, str]]
        ],
        columns: List[ColumnSpec],
        has_header: bool = True,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> None:
        """
        Initializes the fixed_width_file_loader

        :param view: The name of the view that the resultant DataFrame will be stored in
        :param file_path: The path to the fixed width file to load
        :param columns: A list of columns defined using a ColumnSpec that defines the name, start_index, length and DataType for the column
        :param name: sets the name of the transformer as it will appear in logs
        :param parameters:
        :param progress_logger:

        example:
            FrameworkFixedWidthLoader(
                view="my_view",
                file_path=test_file_path,

                columns=[
                    ColumnSpec(column_name="id", start_pos=1, length=3, data_type=StringType()),

                    ColumnSpec(
                        column_name="some_date", start_pos=4, length=8, data_type=StringType()
                    ),

                    ColumnSpec(
                        column_name="some_string",
                        start_pos=12,
                        length=3,
                        data_type=StringType(),
                    ),

                    ColumnSpec(
                        column_name="some_integer",
                        start_pos=15,
                        length=4,
                        data_type=IntegerType(),
                    ),
                ],
            )
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        self.logger: Logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=None)

        self.file_path: Param[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.columns: Param[List[ColumnSpec]] = Param(self, "columns", "")
        self._setDefault(columns=None)

        self.has_header: Param[bool] = Param(self, "has_header", "")
        self._setDefault(has_header=True)

        assert (
            isinstance(file_path, Path)
            or isinstance(file_path, str)
            or callable(file_path)
        ), type(file_path)
        assert file_path

        self.logger.info(f"Received file_path: {file_path}")

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view = self.getView()
        file_path: Union[
            Path, str, Callable[[Optional[str]], Union[Path, str]]
        ] = self.getFilePath()
        if callable(file_path):
            file_path = file_path(self.loop_id)
        columns: List[ColumnSpec] = self.getColumns()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        paths = get_absolute_paths(file_path=file_path)
        has_header = self.getHasHeader()

        if progress_logger:
            progress_logger.write_to_log(f"Loading file for view {view}: {paths}")
            progress_logger.log_param("data_path", ", ".join(paths))

        df_reader: DataFrameReader = df.sql_ctx.read
        df_text = df_reader.text(paths=paths)
        if has_header:
            first: Optional[Row] = df_text.first()
            if first is not None:
                header = first[0]
                df_text = df_text.filter(~col("value").contains(header))
        df_text = df_text.select(
            *[
                trim(col("value").substr(column.start_pos, column.length))
                .cast(column.data_type)
                .alias(column.column_name)
                for column in columns
            ]
        )
        df_text.createOrReplaceTempView(view)

        progress_logger and progress_logger.write_to_log(
            f"Finished Loading file for View[{view}]: {paths}"
        )
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getColumns(self) -> List[ColumnSpec]:
        return self.getOrDefault(self.columns)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getHasHeader(self) -> bool:
        return self.getOrDefault(self.has_header)



from pathlib import Path
from typing import Union, List, Dict, Any, Optional, Callable

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param

from pyspark.sql.dataframe import DataFrame

from pyspark.sql.types import StructType

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger

from spark_pipeline_framework.transformers.framework_local_file_loader.v1.framework_local_file_loader import (
    FrameworkLocalFileLoader,
)
from spark_pipeline_framework.utilities.file_modes import FileReadModes, FileJsonTypes


class FrameworkJsonLoader(FrameworkLocalFileLoader):
    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    @capture_parameters
    def __init__(
        self,
        view: str,
        file_path: Union[
            str, List[str], Path, Callable[[Optional[str]], Union[Path, str]]
        ],
        clean_column_names: bool = False,
        name: Optional[str] = None,
        progress_logger: Optional[ProgressLogger] = None,
        parameters: Optional[Dict[str, Any]] = None,
        limit: Optional[int] = None,
        delimiter: str = ",",
        has_header: bool = True,
        infer_schema: bool = False,
        cache_table: bool = True,
        schema: Optional[StructType] = None,
        create_file_path: bool = False,
        use_schema_from_view: Optional[str] = None,
        json_file_type: Optional[str] = None,
        mode: str = FileReadModes.MODE_PERMISSIVE,
        stream: bool = False,
        delta_lake_table: Optional[str] = None,
    ) -> None:
        """
        class that handles loading of json files


        :param view: view to populate with the data in the file
        :param file_path: where to store the FHIR resources retrieved.  Can be either a path or a function
                    that returns path. Function is passed a loop number if inside a loop or else None
        :param delimiter:
        :param limit:
        :param has_header:
        :param infer_schema:
        :param cache_table: whether to cache the table in memory after loading it
        :param schema: schema to use when loading the data
        :param clean_column_names: whether to remove invalid characters from column names
                                    (i.e., characters that Spark does not like)
        :param create_file_path: whether to create the file path if it does not exist
        :param use_schema_from_view: use schema from this view when loading data
        :param json_file_type: type of json.  None means to autodetect
        :param mode: file loading mode
        :param stream: whether to stream the data
        :param delta_lake_table: store data in this delta lake
        """

        super().__init__(
            view=view,
            file_path=file_path,
            clean_column_names=clean_column_names,
            name=name,
            parameters=parameters,
            progress_logger=progress_logger,
            delimiter=delimiter,
            limit=limit,
            has_header=has_header,
            infer_schema=infer_schema,
            cache_table=cache_table,
            schema=schema,
            create_file_path=create_file_path,
            use_schema_from_view=use_schema_from_view,
            mode=mode,
            stream=stream,
            delta_lake_table=delta_lake_table,
        )

        self.multiLine: Param[bool] = Param(self, "multiLine", "")
        self._setDefault(multiLine=False)

        self.json_file_type: Param[Optional[str]] = Param(self, "json_file_type", "")
        self._setDefault(json_file_type=json_file_type)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def preprocess(self, df: DataFrame, absolute_paths: List[str]) -> None:
        """
        In pre-processing we try to detect whether the file is a normal json or ndjson
        :param df: DataFrame
        :param absolute_paths: list of paths
        """
        json_file_type: Optional[str] = self.getJsonFileType()
        if json_file_type == FileJsonTypes.JSON:
            self.setMultiLine(True)
        elif json_file_type == FileJsonTypes.NDJSON:
            self.setMultiLine(False)
        else:
            assert absolute_paths
            text_df: DataFrame = df.sql_ctx.read.text(absolute_paths)
            # read the first line of the file
            first_line: str = text_df.select("value").limit(1).collect()[0][0]
            if first_line.lstrip().startswith("["):
                self.setMultiLine(True)
            else:
                self.setMultiLine(False)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setMultiLine(self, value: bool) -> "FrameworkJsonLoader":
        self._paramMap[self.multiLine] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMultiLine(self) -> bool:
        return self.getOrDefault(self.multiLine)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderFormat(self) -> str:
        return "json"

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderOptions(self) -> Dict[str, Any]:
        options: Dict[str, Any] = {"multiLine": self.getMultiLine()}
        return options

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getJsonFileType(self) -> Optional[str]:
        return self.getOrDefault(self.json_file_type)



from pathlib import Path

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType

from spark_pipeline_framework.transformers.framework_json_loader.v1.framework_json_loader import (
    FrameworkJsonLoader,
)
from tests.conftest import clean_spark_session


def test_correctly_loads_csv_with_clean_flag_off(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_file_path: str = f"{data_dir.joinpath('column_name_test.json')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    FrameworkJsonLoader(
        view="books", file_path=test_file_path, clean_column_names=False
    ).transform(df)

    # noinspection SqlDialectInspection
    result: DataFrame = spark_session.sql("SELECT * FROM books")

    # Assert
    assert result.count() == 2

    assert result.collect()[1]["title"] == "Essentials of Programming Languages"
    assert len(result.collect()[1]["authors"]) == 2
    assert result.collect()[1]["authors"][0]["surname"] == "Friedman"
    assert (
        result.collect()[1]["Ugly column,with;chars{that}parquet(does)not	like=much_-"]
        == 3
    )



from typing import Dict, Any, Optional, List

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkPartitioner(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        desired_partitions: Optional[int] = None,
        partition_by: Optional[List[str]] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.desired_partitions: Param[Optional[int]] = Param(
            self, "desired_partitions", ""
        )
        self._setDefault(desired_partitions=desired_partitions)

        self.partition_by: Param[Optional[List[str]]] = Param(self, "partition_by", "")
        self._setDefault(partition_by=partition_by)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        desired_partitions: Optional[int] = self.getDesiredPartitions()
        partition_by: Optional[List[str]] = self.getPartitionBy()

        result_df: DataFrame = df.sql_ctx.table(view)
        num_partitions: int = result_df.rdd.getNumPartitions()
        self.logger.info(f"view {view} has {num_partitions} partitions")
        if desired_partitions and partition_by:
            result_df = result_df.repartition(desired_partitions, partition_by[0])
        elif desired_partitions:
            result_df = result_df.repartition(desired_partitions)
        elif partition_by:
            result_df = result_df.repartition(partition_by[0])
        else:
            return df

        num_partitions = result_df.rdd.getNumPartitions()
        self.logger.info(
            f"view {view} has {num_partitions} partitions after repartition"
        )
        result_df.createOrReplaceTempView(view)

        return result_df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDesiredPartitions(self) -> Optional[int]:
        return self.getOrDefault(self.desired_partitions)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getPartitionBy(self) -> Optional[List[str]]:
        return self.getOrDefault(self.partition_by)



from typing import Any, Dict, List, Optional, Union

from mlflow.entities import RunStatus  # type: ignore
from pyspark.ml.base import Transformer
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.transformers.framework_csv_exporter.v1.framework_csv_exporter import (
    FrameworkCsvExporter,
)

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.transformers.framework_validation_transformer.v1.framework_validation_transformer import (
    pipeline_validation_df_name,
)
from spark_pipeline_framework.utilities.FriendlySparkException import (
    FriendlySparkException,
)
from spark_pipeline_framework.utilities.pipeline_helper import create_steps


class FrameworkPipeline(Transformer):
    def __init__(
        self,
        parameters: Dict[str, Any],
        progress_logger: ProgressLogger,
        run_id: Optional[str],
        client_name: Optional[str] = None,
        vendor_name: Optional[str] = None,
        data_lake_path: Optional[str] = None,
        validation_output_path: Optional[str] = None,
    ) -> None:
        """
        Base class for all pipelines
        :param parameters:
        :param progress_logger:
        """
        super(FrameworkPipeline, self).__init__()
        self.transformers: List[Transformer] = []
        self.steps: List[Union[Transformer, List[Transformer]]] = []
        if run_id:
            self._run_id: str = run_id
        if client_name:
            parameters["CLIENT_NAME"] = client_name
            self.client_name: str = client_name
        if vendor_name:
            parameters["VENDOR_NAME"] = vendor_name
            self.vendor_name: str = vendor_name
        if data_lake_path:
            self.data_lake_path: Optional[str] = data_lake_path
        elif "data_lake_path" in parameters:
            self.data_lake_path = parameters["data_lake_path"]

        self.validation_output_path: Optional[str] = validation_output_path

        self.__parameters: Dict[str, Any] = parameters
        self.progress_logger: ProgressLogger = progress_logger

    @property
    def parameters(self) -> Dict[str, Any]:
        return self.__parameters

    @property
    def run_id(self) -> str:
        return self._run_id

    # noinspection PyUnusedLocal
    def fit(self, df: DataFrame) -> "FrameworkPipeline":
        return self

    def _transform(self, df: DataFrame) -> DataFrame:
        try:
            # if steps are defined but not transformers then convert steps to transformers first
            if len(self.steps) > 0 and len(self.transformers) == 0:
                self.transformers = self.create_steps(self.steps)
            # get the logger to use
            logger = get_logger(__name__)
            count_of_transformers: int = len(self.transformers)
            i: int = 0
            pipeline_name: str = self.__class__.__name__

            self.progress_logger.log_event(
                event_name=pipeline_name,
                event_text=f"Starting Pipeline {pipeline_name}",
            )
            self.progress_logger.log_params(params=self.__parameters)

            for transformer in self.transformers:
                assert isinstance(transformer, Transformer), type(transformer)
                # assert isinstance(transformer, FrameworkTransformer), type(transformer)
                try:
                    i += 1
                    logger.info(
                        f"---- Running pipeline [{pipeline_name}] transformer [{transformer}]  "
                        f"({i} of {count_of_transformers}) ----"
                    )
                    self.progress_logger.start_mlflow_run(
                        run_name=str(transformer), is_nested=True
                    )

                    with ProgressLogMetric(
                        progress_logger=self.progress_logger,
                        name=str(transformer) or "unknown",
                    ):
                        self.progress_logger.log_event(
                            pipeline_name,
                            event_text=f"Running pipeline step {transformer}",
                        )
                        df = transformer.transform(dataset=df)
                        self.progress_logger.log_event(
                            pipeline_name,
                            event_text=f"Finished pipeline step {transformer}",
                        )
                    self.progress_logger.end_mlflow_run()
                except Exception as e:
                    if hasattr(transformer, "getName"):
                        # noinspection Mypy
                        stage_name = transformer.getName()
                    else:
                        stage_name = transformer.__class__.__name__
                    logger.error(
                        f"!!!!!!!!!!!!! pipeline [{pipeline_name}] transformer "
                        + "[{stage_name}] threw exception !!!!!!!!!!!!!"
                    )
                    # use exception chaining to add stage name but keep original exception
                    friendly_spark_exception: FriendlySparkException = (
                        FriendlySparkException(exception=e, stage_name=stage_name)
                    )
                    error_messages: List[str] = (
                        friendly_spark_exception.message.split("\n")
                        if friendly_spark_exception.message
                        else []
                    )
                    for error_message in error_messages:
                        logger.error(msg=error_message)

                    if hasattr(transformer, "getSql"):
                        # noinspection Mypy
                        logger.error(transformer.getSql())
                    logger.error(
                        "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
                    )
                    self.progress_logger.log_exception(
                        event_name=pipeline_name,
                        event_text=f"Exception in Stage={stage_name} -- {'-'.join(error_messages)}",
                        ex=e,
                    )
                    self.progress_logger.end_mlflow_run(status=RunStatus.FAILED)
                    raise friendly_spark_exception from e

            self.progress_logger.log_event(
                event_name=pipeline_name,
                event_text=f"Finished Pipeline {pipeline_name}",
            )
            return df
        finally:
            self._check_validation(df)

    def _check_validation(self, df: DataFrame) -> None:
        tables_df = df.sql_ctx.tables().filter(
            f"tableName ='{pipeline_validation_df_name}'"
        )
        if tables_df.count() == 1 and self.validation_output_path:
            FrameworkCsvExporter(
                view=pipeline_validation_df_name,
                file_path=self.validation_output_path,
                header=True,
                parameters=self.parameters,
                progress_logger=self.progress_logger,
            ).transform(df)
            errors_df = df.sql_ctx.sql(
                f"SELECT * from {pipeline_validation_df_name} where is_failed == 1"
            )
            error_count = errors_df.count()
            assert error_count == 0, (
                f"Pipeline failed validation, there were {error_count} errors."
                + " Validation dataframe written to {self.validation_output_path}"
            )

    # noinspection PyMethodMayBeStatic
    def create_steps(
        self,
        my_list: Union[
            List[Transformer],
            List[FrameworkTransformer],
            List[Union[Transformer, List[Transformer]]],
            List[Union[FrameworkTransformer, List[FrameworkTransformer]]],
        ],
    ) -> List[Transformer]:
        return create_steps(my_list)

    def finalize(self) -> None:
        pass



import re
from pathlib import Path
from tempfile import TemporaryDirectory
from types import TracebackType
from typing import Optional, List, Dict, Any

import mlflow  # type: ignore
from mlflow.entities import Experiment, RunStatus  # type: ignore

from spark_pipeline_framework.event_loggers.event_logger import EventLogger
from spark_pipeline_framework.logger.yarn_logger import get_logger


class MlFlowConfig:
    def __init__(
        self,
        mlflow_tracking_url: str,
        artifact_url: str,
        experiment_name: str,
        flow_run_name: str,
        parameters: Dict[str, Any],
    ):
        self.mlflow_tracking_url = mlflow_tracking_url
        self.artifact_url = artifact_url
        self.experiment_name = experiment_name
        self.flow_run_name = flow_run_name
        self.parameters = parameters


class ProgressLogger:
    def __init__(
        self,
        event_loggers: Optional[List[EventLogger]] = None,
        mlflow_config: Optional[MlFlowConfig] = None,
    ) -> None:
        self.logger = get_logger(__name__)
        self.event_loggers: Optional[List[EventLogger]] = event_loggers
        self.mlflow_config: Optional[MlFlowConfig] = mlflow_config

    def __enter__(self) -> "ProgressLogger":
        if self.mlflow_config is None:
            self.logger.info("MLFLOW IS NOT ENABLED")
            return self
        self.logger.info("MLFLOW IS ENABLED")
        mlflow.set_tracking_uri(self.mlflow_config.mlflow_tracking_url)
        self.logger.info(f"MLFLOW TRACKING URL: {mlflow.get_tracking_uri()}")

        # get or create experiment
        experiment: Experiment = mlflow.get_experiment_by_name(
            name=self.mlflow_config.experiment_name
        )

        if experiment is None:
            experiment_id: str = mlflow.create_experiment(
                name=self.mlflow_config.experiment_name,
                artifact_location=self.mlflow_config.artifact_url,
            )
        else:
            experiment_id = experiment.experiment_id
        mlflow.set_experiment(experiment_id=experiment_id)

        mlflow.start_run(run_name=self.mlflow_config.flow_run_name)
        self.logger.info(f"MLFLOW ARTIFACTS URL: {mlflow.get_artifact_uri()}")
        # set the parameters used in the pipeline run
        self.log_params(params=self.mlflow_config.parameters)

        return self

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.logger.info("ENDING PARENT RUN")
        if exc_value:
            # there was an exception so mark the parent run as failed
            self.end_mlflow_run(status=RunStatus.FAILED)
        # safe to call without checking if we have a tracking url set for mlflow
        mlflow.end_run()

    def start_mlflow_run(self, run_name: str, is_nested: bool = True) -> None:
        if self.mlflow_config is None:
            return
        mlflow.start_run(run_name=run_name, nested=is_nested)

    def end_mlflow_run(self, status: RunStatus = RunStatus.FINISHED) -> None:
        mlflow.end_run(status=RunStatus.to_string(status))

    def log_metric(self, name: str, time_diff_in_minutes: float) -> None:
        self.logger.info(f"{name}: {time_diff_in_minutes} min")
        if self.mlflow_config is not None:
            try:
                mlflow.log_metric(
                    key=self.__mlflow_clean_string(name), value=time_diff_in_minutes
                )
            except Exception as e:
                self.log_event("mlflow log metric error", str({e}))

    def log_param(self, key: str, value: str) -> None:
        self.write_to_log(name=key, message=value)
        if self.mlflow_config is not None:
            try:
                mlflow.log_param(
                    key=self.__mlflow_clean_string(key),
                    value=self.__mlflow_clean_param_value(value),
                )
            except Exception as e:
                self.log_event("mlflow log param error", str({e}))

    def log_params(self, params: Dict[str, Any]) -> None:
        if self.mlflow_config is not None:
            # intentionally not using mlflow.log_params due to issues with its SqlAlchemy implementation
            for key, value in params.items():
                self.log_param(key=key, value=value)

    def __mlflow_clean_string(self, value: str) -> str:
        """

        MLFlow keys may only contain alphanumerics, underscores (_),
        dashes (-), periods (.), spaces ( ), and slashes (/)


        mlflow run metric names through https://docs.python.org/3/library/os.path.html#os.path.normpath when validating
        metric names (https://github.com/mlflow/mlflow/blob/217799b10780b22f787137f80f5cf5c2b5cf85b1/mlflow/utils/validation.py#L95).
        one side effect of this is if the value contains `//` it will be changed to `/` and fail the _validate_metric_name check.
        """
        value = str(value).replace("//", "/")
        return re.sub(r"[^\w\-\.\s\/]", "-", value)

    def __mlflow_clean_param_value(self, param_value: str) -> str:
        """
        replace sensitive values in the string with asterisks
        """
        sensitive_value_replacement: str = "*******"
        db_password_regex = r"(?<=:)\w+(?=@)"
        cleaned_value = re.sub(
            db_password_regex, sensitive_value_replacement, str(param_value)
        )

        return cleaned_value

    # noinspection PyUnusedLocal
    def log_artifact(
        self, key: str, contents: str, folder_path: Optional[str] = None
    ) -> None:
        try:
            with TemporaryDirectory() as temp_dir_name:
                data_dir: Path = Path(temp_dir_name)
                file_path: Path = data_dir.joinpath(key)
                with open(file_path, "w") as file:
                    file.write(contents)
                    self.logger.info(f"Wrote sql to {file_path}")

                if self.mlflow_config is not None:
                    mlflow.log_artifact(local_path=str(file_path))

        except Exception as e:
            self.log_event("Error in log_artifact writing to mlflow", str(e))

    def write_to_log(self, name: str, message: str = "") -> bool:
        self.logger.info(name + ": " + str(message))
        return True

    def log_exception(self, event_name: str, event_text: str, ex: Exception) -> None:
        self.log_artifact("_exception.txt", str(ex))
        if self.event_loggers:
            for event_logger in self.event_loggers:
                event_logger.log_exception(
                    event_name=event_name, event_text=event_text, ex=ex
                )

    def log_progress_event(
        self,
        event_name: str,
        current: int,
        total: int,
        event_format_string: str,
        backoff: bool = True,
    ) -> None:
        self.logger.info(event_format_string.format(event_name, current, total))
        if self.event_loggers:
            for event_logger in self.event_loggers:
                event_logger.log_progress_event(
                    event_name=event_name,
                    current=current,
                    total=total,
                    event_format_string=event_format_string,
                    backoff=backoff,
                )

    def log_event(self, event_name: str, event_text: str) -> None:
        self.write_to_log(name=event_name, message=event_text)
        if self.event_loggers:
            for event_logger in self.event_loggers:
                event_logger.log_event(event_name=event_name, event_text=event_text)
